#include "operator.hh"
#include "timers.hh"
#include "waves.hh"

Operator::Operator() : outputlevel(0), target(0), source(0), chainway(mod_none)
{
	envelopetimer = new timer;
	envpos1 = 4;
	envpos2 = 0;
}

Operator::~Operator()
{
	delete envelopetimer;
}

void Operator::FixEnvelope()
{
	sustainlevel = outputlevel*(1 - 1/(sustain+1));
	
	switch(envpos1)
	{
		case 0: /* attack */
			effvol = outputlevel * envpos2 / 15.0;
			break;
		case 1: /* decay */
		{
			double tmp = outputlevel - decay * envpos2 / 15.0;
			if(tmp < 0)tmp = 0;
			effvol = tmp;
			break;
		}
		case 2: /* sustain */
		{
			effvol = sustainlevel;
			break;
		}
		case 3: /* release */
		{
			double tmp = sustainlevel - release * envpos2 / 15.0;
			if(tmp < 0)tmp = 0;
			effvol = tmp;
			break;
		}
	}
}

bool Operator::EnvelopeTick()
{
	switch(envpos1)
	{
		case 0: /* attack */
		{
			/* Would never rise! */
			if(!attack)return false;
			
			/* attack=0: never rise
			 * attack=15: immediate rise
			 */
			 
			envpos2 += attack;
			if(envpos2 < 15)
			{
				effvol = outputlevel * envpos2 / 15.0;
				break;
			}
			
			effvol = outputlevel;
			envpos1 = 1;
			envpos2 = 0;
			/* passthru */
		}
		case 1: /* decay */
		{
			/* Would never drop! */
			if(!decay)return false;
			
			double tmp = outputlevel - decay * envpos2 / 15.0;
			if(tmp < 0)tmp = 0;
			effvol = tmp;
			
			++envpos2;
			
			if(effvol > sustainlevel)
				break;
				
			envpos1 = dosustain ? 2 : 3;
			/* passthru */
		}

		case 2: /* sustain */
		{
			effvol = sustainlevel;
			return false;
		}
		
		case 3: /* release */
		{
			/* Would never drop! */
			if(!release)return false;
			
			double tmp = sustainlevel - release * envpos2 / 15.0;
			
			++envpos2;
			
			if(tmp <= 0)
			{
				envpos1 = 4;
				effvol = 0.0;
				return false;
			}
			
			effvol = tmp;
			break;
		}
		default:
			return false;
	}
	return true;
}

void Operator::Run(double secs)
{
	if(target)target->Run(secs);
	
	/* FIXME --- THIS IS LOTS OF FIXME */	
	unsigned ticks = envelopetimer->Advance(0.01, secs);
	
	while(ticks > 0 && EnvelopeTick())--ticks;
}

void Operator::Modulate(Operator &o)
{
	o.source = this;
	
	chainway = mod_freq;
	target   = &o;
}

void Operator::AddTo(Operator &o)
{
	o.source = this;
	
	chainway = mod_add;
	target   = &o;
}

void Operator::Unlink(void)
{
	target   = 0;
	source   = 0;
	chainway = mod_none;
}

double Operator::multiplyfreq(double freq) const
{
	switch(freqmul)
	{
		case 0: return freq/2;
		case 1: return freq;
		case 11: return freq * 10;
		case 13: return freq * 12;
		case 14: return freq * 15;
	}
	return freq * freqmul;
}

void Operator::Write20(unsigned char value)
{
	tremolo = value & 0x80;
	vibrato = value & 0x40;
	dosustain=value & 0x20;
	KSR =     value & 0x10;
	freqmul = value & 15;
	FixEnvelope();
}

void Operator::Write40(unsigned char value)
{
	KSL = value >> 6;
	
	attenuation = (value & 63) * 0.75;
	
	unsigned attenuation = value & 63;  // 0.75 * attenuation   dB
	unsigned power = 63 - attenuation;
	
	outputlevel = power * power / (63.0 * 63.0);
	
	FixEnvelope();
}

void Operator::Write60(unsigned char value)
{
	attack   = value >> 4;
	decay    = value & 15;
	FixEnvelope();
}
	
void Operator::Write80(unsigned char value)
{
	sustain  = value >> 4;
	release  = value & 15;

	FixEnvelope();
}

unsigned OperatorPortOffset(unsigned channel)
{
	unsigned ofs = 0x100 * (channel / 18);
	register unsigned chan18 = channel % 18;
	return ofs + (chan18%6) + 2 * (chan18/6);
}

unsigned PortOffsetOperator(unsigned offset)
{
	unsigned chan = 18 * (offset / 0x100);
	register unsigned low = offset & 0x1F;
	
	chan = low;
	
	if(low >= 0x08)chan -= 2;
	if(low >= 0x10)chan -= 2;
	
	if(chan >= 36)chan = 36-1;
	
	return chan;
}

static const double PII = 3.141592653589793;

double Operator::Produce(unsigned waveno, double freq,
                         double angle, double shift)
{
	if(effvol < 0.01)return 0.0;
	
	double result = waves[waveno&7] (multiplyfreq(freq) * angle + shift);
	
	result *= effvol;
	
	ownoutput = result;
	
	if(target)
	{
		if(chainway == mod_add)
			result += target->Produce(waveno, freq, angle, 0.0);
		else
			result = target->Produce(waveno, freq, angle, result);
	}
	return result;
}

void Operator::KeyOn()
{
	if(target)target->KeyOn();
	
	envpos1 = envpos2 = 0;
	FixEnvelope();
}

void Operator::KeyOff()
{
	if(target)target->KeyOff();
	envpos1 = 3;
	envpos2 = 0;
}

bool Operator::IsProductive() const
{
	return !source;
}

Operator::chaintype Operator::SourceRelation(const Operator &Source) const
{
	if(source != &Source)return mod_none;
	
	return Source.chainway;
}

Operator::chaintype Operator::TargetRelation() const
{
	return chainway;
}
